package gov.nasa.gsfc.dqss;

import java.io.BufferedInputStream;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.HashMap;

import org.apache.log4j.Logger;

import ncsa.hdf.object.Dataset;


public abstract class Mask {
    
	protected Dataset data;
	protected Dataset[] quality;
	protected Dataset[] reference;
	protected long rows = -1;
	protected long cols = -1;
	protected long layers = 1; // default
	protected long dims4 = -1;
	protected long dims5 = -1;
	
	protected String dataVariable = null;
	private   String stagingDirectory = null;
	private   String statisticsFilename = "stats.dat";
	protected Statistics stats;
	protected int indexoffset = 1;
	protected long pCrit = 3;
	DataQuality dataQuality;
	protected long rank = -1;
	protected int indexOffset = 1;
	protected float FillValue = -9999;
	protected String Algorithm;
	
	
	static Logger logger = Logger.getLogger(Mask.class.getName());
	/*
	 * Most reference this constructor.
	 */
	public  Mask(Dataset data, Dataset[] qual,  String dir, DataFile file) {
		
		this.data = data;
		this.quality = qual;
		this.rank = data.getRank();
	    
	    this.stagingDirectory = dir;
	    dataVariable = data.getName();

	    this.stats = new Statistics(statisticsFilename, dataVariable, file,
	    		qual);
	    data.init();
	    long[] dims = data.getDims();
	    if  (file.FileType == "MODIS" && dims == null) {
			logger.error("This is only for development");
			dims = new long[2];
			dims[0] = 406;
			dims[1] = 270;
			rows = 406;
			cols = 270;
		}
	    if (dims.length >= 3) {
	    	rows = dims[0];
	    	cols = dims[1];
	    	layers = dims[2];
	    }
	    else if (dims.length == 2) {
	    	rows = dims[0];
	    	cols = dims[1];
	    	
	    }
	    else {
	    	logger.warn("Dims.length = " + dims.length);
	    }
	    if (dims.length > 3) {
	    	dims4 = dims[3];
	    	dims5 = dims[4];
	    }
	   
	   
		if (file.crit2num.size() > 0) {
			for (int i = 0; i < file.crit2num.size(); ++i) {
				pCrit = file.crit2num.get(i);
				String qlname = file.qs_names.get(i);
				logger.debug("Setting QualityLevel:" + pCrit + " for " + data.getName() + 
				" of type " + (file.get_qs_types()).get(i)); 
				stats.setQualityLevel(pCrit,i,qlname);
			}
		}
		else {
			logger.error("No criteria extracted from ontology for " + dataVariable);
		}
		
		logger.debug("Dims.length = " + dims.length);
	}
	/*
	 * no references
	 */
	public  Mask(Dataset data, Dataset[] qual, Dataset[] ref,  String dir) {
		
		this.data = data;
		this.quality = qual;
		this.reference = ref;
		this.rank = data.getRank();
	    
	    this.stagingDirectory = dir;
	    dataVariable = data.getName();
	    
	    this.stats = new Statistics(statisticsFilename, dataVariable);
	    long[] dims = data.getDims();
	    if (dims.length >= 3) {
	    	rows = dims[0];
	    	cols = dims[1];
	    	layers = dims[2];
	    }
	    else if (dims.length == 2) {
	    	rows = dims[0];
	    	cols = dims[1];
	    }
	    else {
	    	logger.error("Dims.length = " + dims.length);
	    }
	    if (dims.length >= 2) {
	    	layers = dims[2];
	    }
	    if (dims.length > 3) {
	    	dims4 = dims[3];
	    	dims5 = dims[4];
	    }
		logger.debug("Dims.length = " + dims.length);
	}
	/*
	 * Perhaps param is not necessary as it is in data.getName();
	 * 1 reference
	 */
	public  Mask(Dataset data, Dataset[] qual, 
			 long criteria, String dir) {
		this.data = data;
		this.quality = qual;
		this.layers = data.getRank();
		
		pCrit = criteria;
		this.stagingDirectory = dir;
		dataVariable = data.getName();
		
		this.stats = new Statistics(statisticsFilename, dataVariable);
	    long[] dims = data.getDims();
	    if (dims.length >= 1) {
	    	rows = dims[0];
	    	cols = dims[1];
	    	
	    }
	    else {
	    	logger.error("Dims.length = " + dims.length);
	    }
	    
	    if (dims.length > 2) {
	    	layers = dims[2];
	    }
	    if (dims.length > 3) {
	    	dims4 = dims[3];
	    	dims5 = dims[4];
	    }
		logger.debug("Dims.length = " + dims.length);
	    
	}
	
	
	public  abstract void doMask() throws Exception ;
	public  abstract void doMask_reflect() throws Exception ;
	
	 public boolean writeStats() {
		 String percentage = stats.getPercentage();	 
		 String output = dataVariable  + " mask retained " +
		 stats.kept + " of " + stats.total + " pixels. " + percentage + "%";
		 //String currentcontents = fileRead();
		 return DQSS.fileAppend(output);
	 }
	 
	 private boolean fileAppend(String newStats) {
		 boolean status = false;
		    try{
		        // Create file 
		        FileWriter fstream = new FileWriter(stagingDirectory + File.separator + this.statisticsFilename,true);
		            BufferedWriter out = new BufferedWriter(fstream);
		        out.write(newStats + "\n");
		        //Close the output stream
		        out.close();
		        status = true;
		        }catch (Exception e){//Catch exception if any
		          logger.error("Error: " + e.getMessage());
		        }
		        return status;
	 }
	 
	 private String fileRead() {
		   
		   String lastline = null;
		   try {
			   File file = new File(stagingDirectory + File.separator + this.statisticsFilename);
			   if (file.exists()) {
				   FileInputStream  fis = new FileInputStream(file);
               
			      // Here BufferedInputStream is added for fast reading.
				   BufferedInputStream bis = new BufferedInputStream(fis);
				   DataInputStream dis = new DataInputStream(bis);

			      // dis.available() returns 0 if the file does not have more lines.
			      while (dis.available() != 0) {

			      // this statement reads the line from the file and print it to
			        // the console.
			    	  lastline += dis.readLine();
			       
			        
			      }

			      // dispose all the resources after using them.
			      fis.close();
			      bis.close();
			      dis.close();
			   }
			    } catch (FileNotFoundException e) {
			      e.printStackTrace();
			    } catch (IOException e) {
			      e.printStackTrace();
			    }
			    return lastline;
	   }
	   private boolean fileWrite(String current, String additionalStatsForNewParameter ) {
		   boolean status = false;
		   try{
			    // Create file 
			    FileWriter fstream = new FileWriter(stagingDirectory + File.separator + this.statisticsFilename);
			        BufferedWriter out = new BufferedWriter(fstream);
			    if (current != null) {
			    	out.write(current );
			    	out.newLine();
			    }
			    else  {
			    	logger.debug("Mask::fileWrite current was null");
			    }
			    if (additionalStatsForNewParameter != null) {
			    	out.write(additionalStatsForNewParameter );
			    	out.newLine();
			    }
			    else  {
			    	logger.error("Mask::fileWrite  additionalStatsForNewParameter was null");
			    }
			    //Close the output stream
			    out.close();
			    }catch (Exception e){//Catch exception if any
			      logger.error("Mask::fileWrite Error: " + e.getMessage());
			      return status;
			    }
			    status = true;
			    return status;
	   }
	   
	 public Statistics getStats() {
		 return this.stats;
	 }
	 

	 /*
	  * Endian: byte order, not bit order: Big: byte & 0x0000000E
	  * In Big Endian systems the most significant bit is the 0th or first bit. So in the chart
	  * below the surface type has startbit 0 and stop bit 1.
	  * The surface type flag is bits 0 and 1. The usefulness flag is bit 1
	  * The problem I was having resolving negative numbers (in evaluating the surface type flag)
	  * completely disappeared when I &'ed it with a 4 byte integer. 
	  * 11000101 & 0x000000C0 threw the 8th bit way into the number past any consideration of 
	  * a sign bit.
	  */
	 
	 /* Quality_Assurance_Near_Infrared: Length: 1 byte (8 bits)
	   Flag Name						Number of Bits		Bit Values		Definitions
	   Total Precipitable Water (NIR)	1					0,1				Not useful, Useful (startbit: 7  stopbit:7 )(right most bit)
			   Usefulness flag
		byte & 0x00000001 >> 0;
			   
		Total Precipitable Water (NIRT) 3					0,1,2,3,		Bad(Fill) Marginal,Good,VG (startbit: 4, stopbit: 6)
		byte & 0x0000000E >> 1;
		
		Confidence Flag
		Inversion Method Used (NIR)     2                   0,1,2           2 chan ratio, 3 chan ratio, no retrival (startbit:2, stopbit:3)
		byte & 0x00000030 >> 4;
		
		Surface Type                    2					0,1,2,3			Bright Land, Clear Ocean, Cloud, Sunglint (startbit: 0, stopbit:1)
		byte & 0x000000C0 >> 6;
		
		Note. For the sake of making this programmatic (based on received start and stop bits:
		int  org = (a & 0x0000C0) >> 4;    is same as:  int  usint = (a & 192) >> 4;
             org = (a & 0x0000000E) >> 1;  is same as:  usint = (a & 15) >> 1;
		
	*/
	 /*
	  * So if shifting first:
	  *       startbit:7  stopbit:7 (1 bit)  mask = 0x00000001 (nbits) shift 0 (7-stopbit) (usefulness)
	  *       startbit:4, stopbit:6 (3 bits) mask = 0x00000007         shift 1             (totprecipwater)
	  *       startbit:2, stopbit:3 (2 bits) mask = 0x00000003         shift 4             (inversion)
	  *       startbit:0, stopbit:1 (2 bits) mask = 0x00000003         shift 6             (surface)
	  *       1 -> 1 (2 ^ x - 1)
	  *       2 -> 3
	  *       3 -> 7
	  *       4 -> 15
	  */
	 protected int evaluateQualityFlag(Byte data, int startBit, int stopBit) {
		 int shift = 7 - stopBit; // note this only works if the least significant bit is called bit 7
		 int nbits =  (stopBit - startBit) + 1;
		 int mask  =  (2 ^ nbits) - 1;
		 
		 int preshift = data;
		 int shifted = preshift >> shift;
		 
		 int flag  = shifted & mask;
		 return flag;
	 }

	 protected void testShiftingFirst (Byte qa, Byte cld) {
		 int official = evaluateCloudinessFlag(cld);
		 int preshift = cld;
		 int shifted = preshift >> 6;
		 int cldMask = 0x00000003;
		 int lateshift = shifted & cldMask;
		 if (lateshift != official) {
			 logger.debug ("evaluateCloudinessFlag official:" + official + " preshifted:" + lateshift);
		 }
		 official = evaluateSurfaceBits(qa) ;
		 preshift = qa;
		 shifted = preshift >> 6;
		 cldMask = 0x00000003;
		 lateshift = shifted & cldMask;
		 if (lateshift != official) {
			 logger.debug ("evaluateSurfaceBits official:" + official + " preshifted:" + lateshift);
		 }
		 
		 official = evaluateTotalPrecipConfFlag(qa);
		 preshift = qa;
		 shifted = preshift >> 1;
		 cldMask = 0x00000007;
		 lateshift = shifted & cldMask;
		 if (lateshift != official) {
			 logger.debug ("evaluateTotalPrecipConfFlag official:" + official + " preshifted:" + lateshift);
		 }
		
		 
	 }
	 
	 
	 private int shiftfirstevaluateSurfaceBits(Byte a) {

	        int surfaceMask = 0x000000C0;
			int surface = (a & surfaceMask) >> 6;
	  	    return surface;
		}
	 private int evaluateSurfaceBits(Byte a) {
		 int preshift = a;
		 int shifted = preshift >> 6;
		 int cldMask = 0x00000003;
		 int surface  = shifted & cldMask;
	  	 return surface;
		}
	 protected int shiftlaterevaluateCloudinessFlag(Byte a) {

	        int cldMask = 0x000000C0;
			int cld = (a & cldMask) >> 6;
	  	    return cld;
		}
	 protected int evaluateCloudinessFlag(Byte a) {
		 int preshift = a;
		 int shifted = preshift >> 1;
		 int cld = shifted &  0x00000003;
  	     return cld;
		}
		private int evaluateConfidenceBits(Byte a) {

	        int confidenceMask = 0x0000000E;
			int confidence = (a & confidenceMask) >> 1;
	  	    return confidence;
		}
		protected int evaluateUsefulnessBits(Byte a) {

	        int usefulnessMask = 0x00000001;
			int usefulness = (a & usefulnessMask) ;
	  	    return usefulness;
		}
		protected int shiftfirstevaluateTotalPrecipConfFlag(Byte a) {
			
		    int confidenceMask = 0x0000000E;
		    int confidence = (a & confidenceMask) >> 1;
	  	    return confidence;
		}
		protected int evaluateTotalPrecipConfFlag(Byte a) {
			 int preshift = a;
			 int shifted = preshift >> 1;
			 int confidenceMask = 0x00000007;
		     int confidence = (shifted & confidenceMask);
	  	     return confidence;
		}
		private int evaluateInversionBits(Byte a) {

	        int usefulnessMask = 0x00000030;
			int usefulness = (a & usefulnessMask) >> 4;
	  	    return usefulness;
		}

}


