package gov.nasa.gsfc.dqss;

import java.lang.reflect.Method;
import java.util.HashMap;
import ncsa.hdf.object.Dataset;


/*
import ncsa.hdf.object.Dataset;
<datavar>
<queryOutput>
<QualityView>O3VMRStdView</QualityView>
        <ScreeningVariable>nSurfStd</ScreeningVariable>
        <ScreenVarType>CoordinateConstraintVariable</ScreenVarType>
        <dim/><min/><max/><direction/><Level/>
        <constraint_relationship>lessThan</constraint_relationship>
</queryOutput>
<queryOutput>
<QualityView>O3VMRStdView</QualityView>
        <ScreeningVariable>Qual_O3</ScreeningVariable>
           <ScreenVarType>QualityLevelVariable</ScreenVarType>
           <dim/><min/><max/>
           <direction>negative</direction>
        <Level>0</Level>
        <constraint_relationship/>
</queryOutput>
</datavar>
So this means nSurfStd has a constraint relationship and
Qual_O3 has a direction.
*/

public class Mask1Q1Level1ConMultiLayer extends Mask {
	
	DataFile file;
	String direction = null; 
	public Mask1Q1Level1ConMultiLayer(Dataset data, Dataset[] quality,
			DataFile File, String dir) {
		super(data,quality,dir,File);
		file = File;
		direction = file.qs_direction.get(0);	
	}
	
	 public void doMask_reflect() throws Exception
	  {
		    int indexOffset = 1;
		    short[] qualData = null;
		    int[] qualCon = null;
		    try {
		    	if (file.qs_types.get(0).contains("CoordinateConstraintVariable")) {
		    		qualCon = (int[]) quality[0].read();
		    		qualData  = qualData = (short[]) quality[1].read();
			    	logger.debug("Qual data: " + quality[1].getName() );
			    	logger.debug("Qual constraint: " + quality[0].getName() );
		    	}
		    	else {
		    		qualData = (short[]) quality[0].read();
		    		qualCon  = (int[]) quality[1].read();
			    	logger.debug("Qual data: " + quality[0].getName() );
			    	logger.debug("Qual constraint: " + quality[1].getName() );
		    	}
		    }
		    catch(NullPointerException e) {
		    	logger.error("No screening variable types pulled from ontology");
		    }
		    ReflectDirections rf = new ReflectDirections();
			 //CC
		     Method cc = rf.getMethod(file.qs_coordcon_dir.get(0),Integer.TYPE, Integer.TYPE);
		     Object cclist[] = new Object[2];
		     //quality1 direction
		     Method q1 = rf.getMethod(direction,Short.TYPE, Long.TYPE);
		     Object q1list[] = new Object[2];
		     q1list[1] = pCrit;
	         
	    	long[] start = data.getStartDims(); // get startDims so we can navigate the layers
	    	
	    	float fill = (float) MaskFactory.getFillValue().floatValue();
	    	

	    	for (int layer = 0 ; layer < layers; ++layer) {
	    		logger.debug("Doing layer " + layer + " of " + layers + " for " + data.getName());
	    		if (start.length > 2) { // in case this is used by 2 dim dataset.
	    			start[2] = layer; // this sets the layer outside the class as dims are passed by reference.
	    		}
	    		float[] dataData = (float[]) data.read();
	    		
	    		cclist[1] = layer + indexOffset;
	    		
	    		for (int i = 0; i < rows * cols; ++i) {
                     cclist[0] = qualCon[i];
                     q1list[0] = qualData[i];
                     
	    			 if (dataData[i] == fill) {
	    				 stats.countfill();
	    			 }
	    			 else if ( (Boolean) cc.invoke(rf, cclist)) {
	    				
	    				 if ( (Boolean) q1.invoke(rf, q1list)) {	    
	    					    logger.debug("QC:" + qualData[i] + " <=  Crit:" + pCrit + " so mask pixel = 1");
	    			    		dataData[i] = 1;
			    				stats.inc();
			    			}
			    			else {
			    				logger.debug("QC:" + qualData[i] + " >  Crit:" + pCrit + " so mask pixel = " + fill);
			    				dataData[i] = fill;
			    			}		
	    				    
	    			 }
	    			 else {
	    				 dataData[i] = fill;
	    			 }
	    			
	    		}
	    		data.write(dataData);
	    		
	    	}
	    	stats.setTotal((int) (layers * rows * cols));
	    	
	    }
	  public void doMask() throws Exception
	    {
		    
		    int indexOffset = 1;
		    short[] qualData = null;
		    int[] qualCon = null;
		    try {
		    	if (file.qs_types.get(0).contains("CoordinateConstraintVariable")) {
		    		qualCon = (int[]) quality[0].read();
		    		qualData  = qualData = (short[]) quality[1].read();
			    	logger.debug("Qual data: " + quality[1].getName() );
			    	logger.debug("Qual constraint: " + quality[0].getName() );
		    	}
		    	else {
		    		qualData = (short[]) quality[0].read();
		    		qualCon  = (int[]) quality[1].read();
			    	logger.debug("Qual data: " + quality[0].getName() );
			    	logger.debug("Qual constraint: " + quality[1].getName() );
		    	}
		    }
		    catch(NullPointerException e) {
		    	logger.error("No screening variable types pulled from ontology");
		    }
	    	
	    
	    	long[] start = data.getStartDims(); // get startDims so we can navigate the layers
	    	
	    	float fill = (float) MaskFactory.getFillValue().floatValue();
	    	
	    	logger.debug("length:" + qualData.length);
	    	logger.debug("Fill value:" + fill);
	    	float before;
	    	for (int layer = 0 ; layer < layers; ++layer) {
	    		logger.debug("Doing layer " + layer + " of " + layers + " for " + data.getName());
	    		if (start.length > 2) { // in case this is used by 2 dim dataset.
	    			start[2] = layer; // this sets the layer outside the class as dims are passed by reference.
	    		}
	    		float[] dataData = (float[]) data.read();
	    		
	    		for (int i = 0; i < rows * cols; ++i) {
	    			 before = dataData[i];
	    			 /*
	    			  * What we are doing here is if the users criteria is <= to Qual_*
	    			  * then we keep the data if the layer (+ indexOffset) is >= QualCon
	    			  */
	    			 logger.debug("Layer:"+ layer + "+" + indexOffset + ">=" +qualCon[i]);
	    			 if (dataData[i] == fill) {
	    				 stats.countfill();
	    			 }
	    			 else if (layer + indexOffset >= qualCon[i] ) {
	    				 if ( qualData[i] <= pCrit) {
	    					    
	    					    logger.debug("QC:" + qualData[i] + " <=  Crit:" + pCrit + " so mask pixel = 1");
	    			    		dataData[i] = 1;
			    				stats.inc();
			    			}
			    			else {
			    				logger.debug("QC:" + qualData[i] + " >  Crit:" + pCrit + " so mask pixel = " + fill);
			    				dataData[i] = fill;
			    			}		
	    				    
	    			 }
	    			 else {
	    				 dataData[i] = fill;
	    			 }
	    			
	    		}
	    		data.write(dataData);
	    		
	    	}
	    	stats.setTotal((int) (layers * rows * cols));
	    	
	    }
}
