package gov.nasa.gsfc.dqss;



import gov.nasa.gsfc.Stopwatch;

import java.io.BufferedInputStream;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import ncsa.hdf.hdflib.HDFConstants;
import ncsa.hdf.hdflib.HDFException;
import ncsa.hdf.hdflib.HDFLibrary;
import ncsa.hdf.object.Attribute;
import ncsa.hdf.object.Dataset;
import ncsa.hdf.object.Datatype;
import ncsa.hdf.object.FileFormat;
import ncsa.hdf.object.Group;
import ncsa.hdf.object.HObject;
import ncsa.hdf.object.h4.H4Datatype;
import ncsa.hdf.object.h4.H4File;

import org.apache.log4j.Logger;
/**
 * <p>Title: Reading HDF4 Files with JAVA HDF4 Dataset class</p>
 * <p>Description: This class will:
 * 1. Read in a file.
 * 2. Read in 3 parameters
 * 3. Call the visualizer
 * 4. Apply a simple quality dataset to it's parameter
 * 5. Write out the new datasets
 * 6. Insert the visualized jpeg.
 *  
 * <pre>
 *     "not sure what I should put here yet.
 * </pre>
 * </p>
 *
 * @author Richard Strub GES-DISC
 * @version 1.0
 */
public class Masker 
{
   
	/*
	 * Constructor that does 3D TSurfAirStd and
	 * TAirMWOnlyStd w/Qual_MW_Only_Temp_Strat and Tropo
	 * 
	 * main() is at bottom, only constructor so far is at top
	 * 
	 */
	DataFile file;
	DataQuality dataQuality;
	//DataQualityTransition dataQualityT;
	String stagingDirectory;
	Statistics stats;
	static Logger logger = Logger.getLogger(Masker.class.getName());
	/*
	 * all of the parameters we can handle so far.
	 */
	
	/*
	 * original constructor. Not used after MODAPS Arch
	 */
public void oldMaskerConstructor(String snvid , String param, String criterion, String inputFile, 
		     String dir, String config, String sendToStdOut) throws ClassNotFoundException
    {
    /*
     * One or many data variables:
     */
	List<String> params    = getDataVariables(param);
	List<String> criteria  = getDataVariables(criterion);
	List<Statistics> stats4metadata = new ArrayList<Statistics>();
	/*
	 * Get fill values.
	 * SDProcessing opens the HDF file in a different way and needs
	 * to be closed before other processing proceeds
	 * (and vice versa)
	 * I ended up using a simpler API:
	 * http://www.hdfgroup.org/hdf-java-html/javadocs/ncsa/hdf/object/Dataset.html
	 * rather than the SD (c-like) API:
	 * http://www.hdfgroup.org/hdf-java-html/javadocs/ncsa/hdf/hdflib/HDFLibrary.html#SDsetdimname(int,%20java.lang.String)
	 * ...for most things.
	 * Unfortunately I needed to do a couple of things with SD (HDFLibrary) that
	 * I couldn't do otherwise
	 */
	
		ProductFactory pf = new ProductFactory(params,criteria,stats4metadata,
				dir,config,snvid,inputFile);
		
		/*
		 * original GES-DISC arch  where we get ontology data as we go
		 */
		pf.run();
	
    /*
	 * Update dimension names:
	 * SDProcessing opens the HDF file in a different way and needs
	 * to be opened after the above processing closes the file
	 * (and vice versa)
	 * I ended up using a simpler API:
	 * http://www.hdfgroup.org/hdf-java-html/javadocs/ncsa/hdf/object/Dataset.html
	 * rather than the SD (c-like) API:
	 * http://www.hdfgroup.org/hdf-java-html/javadocs/ncsa/hdf/hdflib/HDFLibrary.html#SDsetdimname(int,%20java.lang.String)
	 * ...for most things.
	 * Unfortunately I needed to do a couple of things with SD (HDFLibrary) that
	 * I couldn't do otherwise
	 */
	Stopwatch s = new Stopwatch();
	s.start();
    SDPostProcessing(inputFile,params,criteria,  pf.FillValues);
    s.stop();
    logger.debug("SDPostProcessing: elapsed time: " + s.getElapsedTime());
    
    if (sendToStdOut != null) {
    	SendToSTDOUT(inputFile);
    }
   
    
  }
/* 
 * Used by OntologyGetter under MODAPS Arch so that it can use
 * getDataVariables...
 */
public Masker() {}
/*
 *
 * MODAPS Arch in the GES-DISC environment where we create the ontology file
 * at Masking time.
 */
public Masker(String snvid , String param, String criterion, String inputFile, 
	     String dir, String config, String sendToStdOut) throws ClassNotFoundException
{
/*
* One or many data variables:
*/
List<String> params    = getDataVariables(param);
List<String> criteria  = getDataVariables(criterion);
List<Statistics> stats4metadata = new ArrayList<Statistics>();
logger.debug("Number of criteria:" + criteria.size());
logger.debug("Number of datavariables:" + params.size());
/*
* Get fill values.
* SDProcessing opens the HDF file in a different way and needs
* to be closed before other processing proceeds
* (and vice versa)
* I ended up using a simpler API:
* http://www.hdfgroup.org/hdf-java-html/javadocs/ncsa/hdf/object/Dataset.html
* rather than the SD (c-like) API:
* http://www.hdfgroup.org/hdf-java-html/javadocs/ncsa/hdf/hdflib/HDFLibrary.html#SDsetdimname(int,%20java.lang.String)
* ...for most things.
* Unfortunately I needed to do a couple of things with SD (HDFLibrary) that
* I couldn't do otherwise
*/
    DQSS.NoOntologyFileProvided=true;
	ProductFactory pf = new ProductFactory(params,criteria,stats4metadata,
			dir,config,snvid,inputFile);
	LabelOntologyFile(inputFile);
	pf.LoadOntologyData();
	/*
	 * original GES-DISC arch  where we get ontology data as we go
	 */
	pf.runWithOntologyFile();

/*
* Update dimension names:
* SDProcessing opens the HDF file in a different way and needs
* to be opened after the above processing closes the file
* (and vice versa)
* I ended up using a simpler API:
* http://www.hdfgroup.org/hdf-java-html/javadocs/ncsa/hdf/object/Dataset.html
* rather than the SD (c-like) API:
* http://www.hdfgroup.org/hdf-java-html/javadocs/ncsa/hdf/hdflib/HDFLibrary.html#SDsetdimname(int,%20java.lang.String)
* ...for most things.
* Unfortunately I needed to do a couple of things with SD (HDFLibrary) that
* I couldn't do otherwise
*/
Stopwatch s = new Stopwatch();
s.start();
//SDPostProcessing(inputFile,params,criteria,  pf.FillValues);
pf.SDPostProcessing(inputFile);
s.stop();
logger.debug("SDPostProcessing: elapsed time: " + s.getElapsedTime());

if (sendToStdOut != null) {
	SendToSTDOUT(inputFile);
}


}

/*
 * MODAPS Arch with a previously written ontology file
 * and ontologyFile in not in default location
 */
public Masker( String inputFile, 
	     String dir, String config, String ontologyFile) throws ClassNotFoundException
{
	List<Statistics> stats4metadata = new ArrayList<Statistics>();

	ProductFactory pf = new ProductFactory(stats4metadata,dir,config,inputFile);
	//pf.LoadOntologyData();
	
	/*
	 * MODAPS arch -  use ontology file just produced.
	 */
   //LabelOntologyFile(inputFile);	
	DQSS.OntFile = ontologyFile;
    pf.runWithOntologyFile();



Stopwatch s = new Stopwatch();
s.start();

/*
* SDPostProcessing has to be after file has been H4File.closed()
*/
pf.SDPostProcessing(inputFile);
pf.SetSpecificAttributesInMask(inputFile);
s.stop();
logger.debug("SDPostProcessing: elapsed time: " + s.getElapsedTime());

}
/*
 * MODAPS Arch with a previously written ontology file
 * and ontologyFile in default location
 */
public Masker( String inputFile, String dir, String config) throws ClassNotFoundException
{
	List<Statistics> stats4metadata = new ArrayList<Statistics>();

	ProductFactory pf = new ProductFactory(stats4metadata,dir,config,inputFile);
	//pf.LoadOntologyData();
	
	/*
	 * MODAPS arch -  use ontology file just produced.
	 */
 
   LabelOntologyFile(inputFile);	
   // DQSS.OntFile = DQSS.stagingDirectory + File.separator + DQSS.OntFile;
   // DQSS.OntFile = DQSS.OntFile;
     
   pf.runWithOntologyFile();


/*
* SDPostProcessing has to be after file has been H4File.closed()
*/
  pf.SDPostProcessing(inputFile); 
  pf.SetSpecificAttributesInMask(inputFile);
}

/* 
 * Update the dimension names of all newly created datasets:
 * dataVariableMask
 * dataVariableOriginal
 */
void LabelOntologyFile(String inputfile) {
	String hdf = new File(inputfile).getName();
	int i = hdf.lastIndexOf(".");
	String ext = hdf.substring(i+1,hdf.length());
	hdf = hdf.replace(ext, "xml");
	DQSS.OntFile =   DQSS.stagingDirectory + File.separator + "DQSS_" + hdf;
	logger.debug("Ontology file is:" + DQSS.OntFile);
}


void WriteAlgorithmsWithDataFile (List<Statistics> stats4metadata, DataFile file)  {
	String AlgStats = "\n";
	String AlgAttribute = "Algorithms Used";
	 Iterator statsitr = stats4metadata.iterator();
	 while(statsitr.hasNext()) {
		 Statistics nextstats = (Statistics) statsitr.next();
		 if (nextstats.getAlgorithm() != null) {
			 AlgStats += nextstats.getVariable() + ": " + nextstats.getAlgorithm() + "\n";
		 }
		 else {
			 logger.warn("No algorithm retrieved for " + nextstats.getVariable());
		 }
	 }
	 Attribute stats_attr;
	try {
		stats_attr = file.getYourAttribute(AlgAttribute,AlgStats.length() );
		
		String[] StatsArray = {AlgStats};
		file.writeStatsToMetadata(stats_attr,StatsArray);
	} catch (Exception e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}
	 
			
}

void SDPostProcessing(String inputFile, List<String> params, List<String> criteria,
		 List<DQSSNumber> FillValues)
    {
	
    try {
	    SDDataFile sid = new SDDataFile(inputFile);
	    Iterator itr = params.iterator();
	    Iterator crititr = criteria.iterator();
	    Iterator fvitr   = FillValues.iterator();
	    
	    while(itr.hasNext()) {
            String name = (String) itr.next();
            /*
             * fill in dimension names for newly created masks
             */
            String criterion = (String) crititr.next();
            DQSSNumber fillvalue = (DQSSNumber) fvitr.next();
            if (! criterion.contains(DQSS.NoScreeningString)) {
                String[] newdimnames = sid.GetDimensionNames(name);
                sid.UpdateDimensionNames(name + DQSS.MaskName, newdimnames);
            /*
             * Set the quality level attribute
             */
                sid.SetAttribute(name + DQSS.MaskName, "Quality", criterion);
                String[] check = sid.GetDimensionNames(name + DQSS.MaskName);
                sid.SetFill(name + DQSS.MaskName, fillvalue);
            }
            else {
                logger.warn("Criteria is " + criterion + " for " + name + " so we are not post-processing it");
            }
        }
	    sid.destroy();
    }
    catch (HDFException e) {
    	e.printStackTrace();
    	logger.error("SD stuff failed");
    }
    
}


public List<String> getDataVariables(String param) {
	
	List<String> params = new ArrayList<String>();   
	StringTokenizer paramsT = null;
	
	
	if (param.contains(",")) {
		paramsT = new StringTokenizer(param,",");
	
		while (paramsT.hasMoreTokens()) {
			params.add(paramsT.nextToken());
		}
	}
	else {
		params.add(param); 
	}
	logger.debug("We found " + params.size() + " params");
	return params;
}
public List<String> getCriteria(String criterion) {
	List<String> criteria = new ArrayList<String>();
	StringTokenizer critT = null;
	if (criterion.contains(",")) {
		
    	critT = new StringTokenizer(criterion,",");
    	
    	while (critT.hasMoreTokens()) {
      	  criteria.add(critT.nextToken());
      	}
    	
	}
	else {
		criteria.add(criterion);
	}
	return criteria;
}
public String[] GetDimensionNames(String SDFileName,String name ,long[] dims) 
{
	
   
    String[] dimnames = new String[dims.length];
    
    boolean status;
	try {
	    int sd_id = HDFLibrary.SDstart(SDFileName, HDFConstants.DFACC_WRITE);
		
		int some_id = HDFLibrary.SDnametoindex(sd_id, name);
		int  sds_id = HDFLibrary.SDselect(sd_id, some_id);
		  doGetInfo(sds_id);
		  for (int dimindex=0; dimindex < dims.length; ++dimindex) {
			
		  	int dim_id = HDFLibrary.SDgetdimid (sds_id, dimindex);
	  		if (dim_id > 0) {
		  		String[] dimname = {""};
		  		int[] args = new int[3];
		  		HDFLibrary.SDdiminfo(dim_id, dimname, args);
		  		logger.debug("dimname:" + dimname[0] + ", " + dim_id);
		  	    dimnames[dimindex] =  dimname[0];
	  	    
	  		}
	  		else {
	  			logger.debug("SDgetdimid("+sds_id+","+dimindex+") failed");
	  		}
	  	
		  		
		  }
		    status = HDFLibrary.SDendaccess(sds_id);
	  		status = HDFLibrary.SDend(sd_id);
		} catch (HDFException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return dimnames;
}

private void doGetInfo(int sds_id) throws HDFException {
	 String[] sdsname={""}; 
	 int[] sizes = new int[3];
     int[] args = new int[3];
	 HDFLibrary.SDgetinfo(sds_id, sdsname, sizes, args);
	 for (int i = 0; i < 3; ++i) {
		 logger.debug(sdsname[0] + " dimsize:" + sizes[i]);
		
	 }
	 
	 logger.debug(sdsname[0] + " rank:    "     + args[0]);
	 logger.debug(sdsname[0] + " datatype:"     + args[1]);
	 logger.debug(sdsname[0] + " nAttributes:"     + args[2]);
}
public void UpdateDimensionNames(String SDFileName, String name,long[] dims,
		String[] newnames)
{
	
 try {
    int sd_id = HDFLibrary.SDstart(SDFileName, HDFConstants.DFACC_WRITE);
	
	int some_id = HDFLibrary.SDnametoindex(sd_id, name);
	int  sds_id = HDFLibrary.SDselect(sd_id, some_id);
	boolean status = false;
	  for (int dimindex=0; dimindex < dims.length; ++dimindex) {
	  	int dim_id = HDFLibrary.SDgetdimid (sds_id, dimindex);
	  	status = HDFLibrary.SDsetdimname (dim_id, newnames[dimindex] + dimindex );
	  	if (status) { 
	  		String[] dimname = {""};
	  		int[] args = new int[3];
	  		HDFLibrary.SDdiminfo(dim_id, dimname, args);
	  		logger.debug("I set dimname to:" + dimname[0]);
	  	    String[] info = {"","","",""};
	  		status = HDFLibrary.SDgetdimstrs(sds_id,info,20);
	  		if (status) {
	  			for (int i = 0; i < 4; ++i) {
	  				logger.debug("dimstrs:" + info[i]);
	  			}
	  		}
	  		else {
	  			logger.error("SDgetdimstrs failed");
	  		}
	  	
	  	}
	  	else {
	  		logger.error("SDsetdimname failed");
	  	}
	    
	  }
	  status = HDFLibrary.SDendaccess(sds_id);
	  status = HDFLibrary.SDend(sd_id);
	} catch (HDFException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}
  }



	
 /*   
     
    private void determineAndPerformAlgorithm(String parameter,DataFile file, DataQuality dataQuality) {
        
        
        	Group newg  = null;
            Group currg = file.getGroup("mask",file.topGroup);
            if (currg != null) {
            	if (currg.getName().equals("mask")) {
            		newg = currg;
            		logger.warn("Using existing group:mask");
            	}
            	else {
            		newg = file.createGroup(parameter);
            		logger.warn("Using NEW group:" + newg.getName() );
            	}
            }
            
            dataQuality.XML2Castor();
            int nQSvars = dataQuality.rootCastor.getQueryOutputCount();
    		
    		String[] qs_names = new String[nQSvars];
    		String[] dims    = new String[nQSvars];
    		long[] mins    = new long[nQSvars];
    		long[] maxs    = new long[nQSvars];
    	    QueryOutput[] qo = dataQuality.rootCastor.getQueryOutput();
    	    for (int i = 0; i < nQSvars; ++i) {
    	    	qs_names[i] = qo[i].getScreeningVariable();
    	    	dims[i]    = qo[i].getDim();
    	    	mins[i]    = qo[i].getMin();
    	    	maxs[i]    = qo[i].getMax();
    	    	
    	    }
    	
        	if (parameter.contains("TAirStd")) {
        		
        	    
        	    Dataset[] qs_var   = file.getQSVars(qs_names);
        	    if (qs_var[0] != null) {
        	    	logger.debug("I Found qs_var:" + qs_var[0].getName());
        	    }
        	    else {
        	    	logger.debug("I could not find a qs_var");
        	    }
        	    String newname = new String("masked" + parameter);
        	    Dataset dtairstd =   file.copyDataset(parameter,newg );
            	TAirStd_3D example = new TAirStd_3D();
               	example.doThisMask(dtairstd,qs_var);
               	example.writeStats();
            }
        	
        
            if (parameter.contains("TAirMWOnly")) {
            	          	
                           
                Dataset[] qs_var   = file.getQSVars(qs_names);
                String newname = new String("masked" + parameter);
        	    Dataset dtmwonly =   file.copyDataset(parameter,newg );
        	    
        	    logger.debug("Doing TwoQVars");
            	TwoQVars example = new TwoQVars(dataQuality.getCriteria() , qs_names, dims, mins,maxs);
               	example.doThisMask(dtmwonly,qs_var);
            }
            
            if (parameter.equals("CH4_total_column")) {
        		
        	    
        	    Dataset[] qs_var   = file.getQSVars(qs_names);
        	    if (qs_var[0] != null) {
        	    	logger.debug("I Found qs_var:" + qs_var[0].getName());
        	    }
        	    else {
        	    	logger.debug("I could not find a qs_var");
        	    }
        	    String newname = new String("masked" + parameter);
        	    Dataset ch4 =   file.copyDataset(parameter,newg );
        	    CH4_2D example = new CH4_2D(dataQuality.getCriteria());
               	example.doThisMask(ch4,qs_var);
            }
        }
    
    */
    
    /*
     * This is something I wrote to see where the mask turned up.
     */
    private void checkLocationOfNewDatasets(String newDataset, Group TopGroup)
    {
    	
    	Dataset d = null;
    	d = getDatasetRecurse(newDataset,TopGroup);
    	if (d != null) {
    		logger.debug("Path of current dataset:" + d.getPath());
    	}
    	
    	//logger.debug("Dataset " + newDataset + " is in group:" + currentGroup.getName());
    }
    
  
    
    
  
  
  
   
    
  /*
   * Critical in the path of getting the quality datasets.  
   */
  private Dataset[] getQSVars(String dataVariable, String[] qs,Group g, H4File newFile ,Group mask) {
    	 
    	Dataset[] qsDatasets = new Dataset[qs.length]; 
    	
    	for (int i = 0; i < qs.length; i++) {
    	    	   logger.debug("Doing:" + qs[i]);
    	   qsDatasets[i] = getDataset(qs[i],g,newFile, mask);
    	   logger.debug("getQSVars:" + ((Dataset) qsDatasets[i]).getName());
    	}
    	
    	
    	return (Dataset[]) qsDatasets;
  }
  
  /*
   * critical in the path of getting and creating the 'data' datasets.
   * 1. Use group to find address of dataset
   * 2. Use createDataset() and populateDataset() to create a copy of the
   * 'data' dataset. This will be the mask.
   * Eventually we want to:
   * a) rename the 'data' dataset.
   * b) create 2 new datasets: 1 for the mask , 1 for the screened dataset.
   * (Lots of other stuff too but I expect these things to be done at this point)
   */
 
  
  
 
 /*
  * This is all that is left after I fixed things. might just put it in getQSVars()
  */
private Dataset getDataset(String datasetName , Group TopGroup, H4File newFile, Group mask)
{
	Dataset d =  getDatasetRecurse(datasetName,TopGroup);
	
	
	return d;
}

  /*
   * Critical in the path of locating datasets.
   *  This needs to be recursive as there are groups within groups
   */
private Dataset getDatasetRecurse(String datasetName , Group TopGroup)
{
	List members = TopGroup.getMemberList();

    int n = members.size();
    HObject obj = null;
    int objectid; 
    Dataset looking = null;
    Dataset d = null;
  
    for (int i=0; i< members.size(); i++)
    {
    	  if (d != null && d instanceof Dataset) {
    	    	logger.debug("Returning:" + d.getName());
    	    	return d;
    	    }
        obj = (HObject)members.get(i);

        /*
         * Each Dataset (SDS) is a member of a Group
         * In the AIRS file I worked with the groups were:
         * L2_Standard_atmospheric&surface_product
         * Geolocation Fields
         * SDS Fields
         * Swath Attributes
         */
        if (obj instanceof Dataset)
        {
        	looking = (Dataset) obj;
        	
        	if (looking.getName().equals(datasetName)){
        		d = looking;
        		logger.debug("Found Dataset:" + d.getName());
        		logger.debug("Found Path:" + d.getPath());
        		
        		return (Dataset) obj;
        	}
        }
        if (obj instanceof Group && d == null)
        {
        	//currentGroup =  (Group) obj;
            d = getDatasetRecurse(datasetName,(Group)obj);
            if (obj instanceof Group ) {
            	logger.debug("Current Group:" + obj.getName() );
            }
                       
        }
        
        
    }

   return d;
}


private void checkdims(long[] dims) {
	 for (int i = 0; i < dims.length; ++i) {
		 logger.debug("Dims:" + dims[i]);
	 }
}

/*
 * Finds the group which the 'data' dataset is in.
 * I am trying to put the new dataset into the same group as the old one.
 */
private Group getGroup(String datasetName , Group currentGroup)
{
	List members = currentGroup.getMemberList();

    int n = members.size();
    HObject obj = null;
    int objectid; 
    Dataset looking = null;
    Dataset d = null;
  
    for (int i=0; i< members.size(); i++)
    {
    	  if (d != null && d instanceof Dataset) {
    	    	logger.debug("Returning:" + d.getName());
    	    	return currentGroup;
    	    }
        obj = (HObject)members.get(i);

       
        if (obj instanceof Dataset)
        {
        	looking = (Dataset) obj;
        	
        	if (looking.getName().equals(datasetName)){
        		d = looking;
        		logger.debug("Found Dataset:" + d.getName());
        		logger.debug("Found Path:" + d.getPath());
        		
        		return currentGroup;
        	}
        }
        if (obj instanceof Group && d == null)
        {
            currentGroup = getGroup(datasetName,(Group)obj);
            if (obj instanceof Group ) {
            	logger.debug("Current Group:" + obj.getName() );
            }
                       
        }
        
        
    }

   return currentGroup;
}


    /*
     * All of the methods below (except for main() )  is leftover from my proof of concept code.
     * Some of the subroutines I actually use - determineDataType, determineDimension
     * Some I use for debugging = displayValues...
     */
    private static void printGroup(Group g, FileFormat testFile) throws Exception
    {
        if (g == null)
            return;

        List members = g.getMemberList();

        int n = members.size();
        HObject obj = null;
        int objectid; 
        for (int i=0; i< members.size(); i++)
        {
            obj = (HObject)members.get(i);
               
            /*
             * Each Dataset (SDS) is a member of a Group
             * In the AIRS file I worked with the groups were:
             * L2_Standard_atmospheric&surface_product
	     * Geolocation Fields
	     * SDS Fields
             * Swath Attributes
             */
            if (obj instanceof Dataset)
            {
               //ReadAndDisplaySDS(obj,testFile);
            }
            else {
                System.out.println("Not trying to print Group:"+obj.getName());
            }
            if (obj instanceof Group)
            {
                printGroup((Group)obj,testFile);
            }
        }
    }

    /**
     * create the file and add groups ans dataset into the file,
     * which is the same as javaExample.H4DatasetCreate
     * @see javaExample.H4DatasetCreate
     * @throws Exception
     */

    private static boolean listable(Dataset dataset) 
    {
       if (dataset.hasAttribute()) { 
          return true;
       }
       return false;
    }

    private static void determineDatatype(Dataset dataset, HashMap sizes) throws Exception
    {
            Datatype Type = dataset.getDatatype();
            H4Datatype HType = (H4Datatype)dataset.getDatatype();
            System.out.println("datatype description:" + HType.getDatatypeDescription());
            //System.out.println("dataset type size:" +  HType.getDatatypeSize()); 
            if (HType.isUnsigned()) {
                //System.out.println("dataset " + dataset.getName() + " is unsigned"); 
            }
            else {
                //System.out.println("dataset " + dataset.getName() + " returns that it is NOT unsigned"); 
            }
            sizes.put("size", HType.getDatatypeSize());
            sizes.put("class",  HType.getDatatypeClass());
            sizes.put("name", dataset.getName());
            if (HType.getDatatypeDescription().contains("float") ) {
                 if (HType.getDatatypeSize() == 4) {
                     sizes.put("type", "FLOAT32");
                 }
                 else if (HType.getDatatypeSize() == 8) {
                     sizes.put("type", "FLOAT64");
                 }
                 else {
                     sizes.put("type", "unknown");
                 }
            }
            if (HType.getDatatypeDescription().contains("integer") ) {
                 if (HType.getDatatypeSize() == 2) {
                     sizes.put("type", "INT16");
                 }
                 else if (HType.getDatatypeSize() == 4) {
                     sizes.put("type", "INT32");
                 }
                 else {
                     sizes.put("type", "unknown");
                 }
            }
                      
    }
 /*
    private static void ReadAndDisplaySDS(HObject obj, FileFormat ff) throws Exception
    {
    
       
        Dataset dataset = (Dataset)obj;
       
        if (listable(dataset)){
            HashMap dimension = new HashMap();  
            determineDatatype(dataset,dimension);
 
            determineDimensions(dataset,dimension);
 
            int[] dims = (int[])dimension.get("dims");
            int x = dims[0];    
            int y = dims[1];    
            int z = -1;
            if (dims.length > 2) {
                z = dims[2];
            }

            System.out.println("SDS:Name:"+dimension.get("name"));
            System.out.println("Type:"+dimension.get("type"));

               if (dimension.get("type") == "INT16") {
        	   short[] dataRead = (short[])dataset.read();
                   //displayValues(x,y,dataRead);
                   displayMinMax(x,y,dataRead);
               }
               if (dimension.get("type") == "INT32") {
        	   int[] dataRead = (int[])dataset.read();
                   //displayValues(x,y,dataRead);
                   displayMinMax(x,y,dataRead);
               }
               if (dimension.get("type") == "FLOAT32") {
       
                   if (dims.length == 3) {
                      //display_ncdumpstyle(dataset,x,y,z); 
                       //displayValues(x,y,z,dataset);
                       displayMinMax(x,y,z,dataset);
                    }
                    else { 
        	       float[] dataRead = (float[])dataset.read();
                      // displayValues(x,y,dataRead);
                       displayMinMax(x,y,dataRead);
                    }
               }
               if (dimension.get("type") == "FLOAT64") {
        	   double[] dataRead = (double[])dataset.read();
                   //displayValues(x,y,dataRead);
                   displayMinMax(x,y,dataRead);
               }
          }
    }


     private void display_ncdumpstyle(Dataset dataset, int x, int y, int z) throws Exception
     {
                     int rank = dataset.getRank();   // number of dimension of the dataset
		     long[] dims2 = dataset.getDims(); // the dimension sizes of the dataset
		     long[] selected = dataset.getSelectedDims(); // the selected size of the dataet
		     long[] start = dataset.getStartDims(); // the off set of the selection
		     long[] stride = dataset.getStride(); // the stride of the dataset
		     int[]  selectedIndex = dataset.getSelectedIndex(); // the selected dimensions for display

		     // select dim1 and dim2 as 2D data for display,and slice through dim0
                     start[0] = 0;  
                     start[1] = 0;  
                     start[2] = 2;  
                     System.err.println("start:" + start[0] + " " +start[1] + " " + start[2]);
                     // Dataset.read() reads orthogonally to ncdump. I'll know better when I ask
                     // Young-In about what ncdump.
                     float[][]dataRead = new float[z][x*y];  
                     for (int startdim = 0; startdim < z; ++startdim) {
			    // when dataset.read() is called, the slection above will be used since
			    // the dimension arrays is passed by reference. Changes of these arrays
			    // outside the dataset object directly change the values of these array
			    // in the dataset object.
                         start[2] = startdim;
        	         float[] data = (float[])dataset.read();
                         for (int i = 0; i < data.length; ++i) {
        	           dataRead[startdim][i] = data[i];
                           if (i == 0 ) {
                             System.err.print(data[i]+ ",");
                           }
                         }
                     }
                     for (int c = 0; c < x*y; ++c) {
                       for (int w = 0; w < z; ++w) {
                             System.out.print(dataRead[w][c]+ ",");
                       }
                       System.out.println();
                     }
} 
   private static void displayValues(int x, int y, short[] dataRead)
   {
	   for (int i=0; i<y; i++) {
		System.out.print("\n"+dataRead[i*x]);
		for (int j=1; j<x; j++) {
			System.out.print(", "+dataRead[i*x+j]);
		}
	   }

    }
   private static void displayValues(int x, int y, int[] dataRead)
   {
	   for (int i=0; i<y; i++) {
		System.out.print("\n"+dataRead[i*x]);
		for (int j=1; j<x; j++) {
			System.out.print(", "+dataRead[i*x+j]);
		}
	   }

    }
   private static void olddisplayValues(int x, int y, float[] dataRead)
   {
	   for (int i=0; i<y; i++) {
		System.out.print("\n"+dataRead[i*x]);
		for (int j=1; j<x; j++) {
			System.out.print("; "+dataRead[i*x+j]);
		}
	   }

    }
   private static void displayValues(int x, int y, float[] dataRead)
   {
	   for (int i=0; i< x * y; i++) {
               
                 System.out.print(dataRead[i] + ", ");
		if (i % 5 == 0 ) {
                  System.out.print("\n");
                }
	   }

    }
 
   private static void displayValues(int x, int y, int z, Dataset dataset) throws Exception
   {
	   long[] start = dataset.getStartDims(); // the off set of the selection
           for(int layer = 0; layer < z; ++layer) {
                start[2] = layer; // get to next layer
	   	   		for (int i=0; i < x*y ; i++) {
        	        float[] dataRead = (float[])dataset.read();
        	        if (dataRead[i] > 0) {
                 	    System.out.print(dataRead[i] + ", ");
        	        }
		 	    	if (i % (x*y - 1) == 0 ) {
                  		System.out.print("\n layer " + layer);
                	}
	   			}
           } 

    }
   
   private static void displayValues(int x, int y, double[] dataRead)
   {
           int f = 1;
	   for (int i=0; i< x * y; i++) {
               
                 System.out.print(dataRead[i] + ", ");
		if (f % 4 == 0 ) {
                  System.out.print("\n");
                }
		if ((f) % 30 == 0 ) {
                  System.out.print("\n\n");
                  f = 0;
                }
                ++f;
	   }

    }

   private static void displayMinMax(int x, int y,short[] dataRead) {
	   if (dataRead.length > 1) {
		
           short max = dataRead[0];
           short min  = dataRead[0];
           for (int i=0; i<y; i++) {
        	   for (int j=1; j<x; j++) {
                   if (dataRead[i*x+j] < min) {
                       min = dataRead[i*x+j];
                   }
                   if (dataRead[i*x+j] > max) {
                       max = dataRead[i*x+j];
                   }
                }
           }
           System.out.println("Min:"+min+" Max:"+max);
	  
	   }
   }
   private static void displayMinMax(int x, int y,int[] dataRead) {
           int max = dataRead[0];
           int min  = dataRead[0];
	   for (int i=0; i<y; i++) {
		for (int j=1; j<x; j++) {
                   if (dataRead[i*x+j] < min) {
                       min = dataRead[i*x+j];
                   }
                   if (dataRead[i*x+j] > max) {
                       max = dataRead[i*x+j];
                   }
                }
           }
	System.out.println("Min:"+min+" Max:"+max);

   }
   private static void displayMinMax(int x, int y,float[] dataRead) {
           float max =  -1000;
           float min  = 1000;
	   for (int i=0; i<y; i++) {
		for (int j=1; j<x; j++) {
                   if (dataRead[i*x+j] < min) {
                       min = dataRead[i*x+j];
                   }
                   if (dataRead[i*x+j] > max) {
                       max = dataRead[i*x+j];
                   }
                }
           }
	System.out.println("Min:"+min+" Max:"+max);

   }
   private static void displayMinMax(int x, int y, int z, Dataset dataset) throws Exception
   {
       float max =  -9999;
       float min  = 10000;
	   long[] start = dataset.getStartDims(); // the off set of the selection
           int f = 1;
           for(int layer = 0; layer < z; ++layer) {
                start[2] = layer; // get to next layer
        	    float[] dataRead = (float[])dataset.read();
	   	   		for (int i=0; i < x*y ; i++) {
                   if (dataRead[i] < min) {
                       min = dataRead[i];
                   }
                   if (dataRead[i] > max) {
                       max = dataRead[i];
                   }
	   			}
           } 
	       System.out.println("Min:"+min+" Max:"+max);

    }

   private static void displayMinMax(int x, int y,double[] dataRead) {
           double max = dataRead[0];
           double min  = dataRead[0];
	   for (int i=0; i<y; i++) {
		for (int j=1; j<x; j++) {
                   if (dataRead[i*x+j] < min) {
                       min = dataRead[i*x+j];
                   }
                   if (dataRead[i*x+j] > max) {
                       max = dataRead[i*x+j];
                   }
                }
           }
	System.out.println("Min:"+min+" Max:"+max);

   }
   
   */
   
   public static void argumentsCheck(String args[]) {
	   
	   
	   for (int i = 0 ; i < args.length; ++i ) {
		   logger.error(args[i]);
	   }
   }
   public static void main( String args[] ) 
   {
   
   	String inputFile = null;
   
   	try { 
   		/*
   		 * MODAPS arch default OntFile location
   		 */
       if (args.length == 3) {
    	           inputFile = args[0];
    	   String stagingDir = args[1];
    	   String config     = args[2];
    	   Masker inst = new Masker(inputFile, stagingDir, config);
       }
       else if (args.length == 4) {
           	       inputFile = args[0];
           String stagingDir = args[1];
           String config     = args[2];
           String ontologyfile=args[3];
           Masker inst = new Masker(inputFile, stagingDir, config, ontologyfile);
       }
       /*
        * Original GES-DISC Testing and Production
        * (We only send to stdout in screener)
        */
       else if (args.length == 6) {
    	    inputFile = args[0];
    	    
    	   String snvid        = args[1];
           String datavariable = args[2];
    	   String criteria     = args[3];
    	   String stagingDir   = args[4];
    	   String config       = args[5];
           Masker inst = new Masker(snvid,datavariable,criteria,inputFile,
        		                    stagingDir,config,null);
       }
       /*
        * 
        */
       else if (args.length == 7) {
    	   inputFile           = args[0];
    	   
           String snvid        = args[1];
           String datavariable = args[2];
    	   String criteria     = args[3];
    	   String stagingDir   = args[4];
    	   String config       = args[5];
    	   String sendToStdOut = args[6];
           Masker inst = new Masker(snvid, datavariable,criteria,inputFile,
        		   stagingDir,config,sendToStdOut);
       }
      
       else {
    	   argumentsCheck(args);
    	   System.err.println("Proper number of arguments not received.");
    	   System.exit(1);
       }
   	}
   	catch (ClassNotFoundException e) {
   		System.err.println("Could not find needed jar files.");
   		e.printStackTrace();
   		System.err.println("Could not find needed jar files.");
   	   	}
       
   	
   }  // end main
   
   public static void SendToSTDOUT(String maskedFile) {
       int status;
       int BUFFERSIZE = 4096*16;
       
       File file = new File(maskedFile);
	   FileInputStream fis = null;
	try {
		fis = new FileInputStream(maskedFile);
	} catch (FileNotFoundException e) {
		// TODO Auto-generated catch block
		e.printStackTrace();
	}
       
       byte[] data = new byte[BUFFERSIZE];
       String sample = data.toString();
       int count;    
       
       try {
    	   while ( (count =   fis.read( data )) != -1 ) {
		           System.out.write( data, 0, count );
		           
		   }
    	   fis.close();
       } catch (IOException e) {
		// TODO Auto-generated catch block
		// probably just EOFe.printStackTrace();
       }
        
        
       
       
       /*
	   try {
		   File file = new File(maskedFile);
		   FileInputStream  fis = new FileInputStream(file);

		      // Here BufferedInputStream is added for fast reading.
		   BufferedInputStream bis = new BufferedInputStream(fis);
		   DataInputStream dis = new DataInputStream(bis);

		      // dis.available() returns 0 if the file does not have more lines.
		      while (dis.available() != 0) {

		      // this statement reads the line from the file and print it to
		        // the console.
		    	  status = dis.readByte();
		        System.out.write(dis.);
		        
		      }

		      // dispose all the resources after using them.
		      fis.close();
		      bis.close();
		      dis.close();

		    } catch (FileNotFoundException e) {
		      e.printStackTrace();
		    } catch (IOException e) {
		    	// EOF
		      //e.printStackTrace();
		    }
		   */
		    
   }
   private String fileRead(String filepath) {
	   
	   String lastline = null;
	   try {
		   File file = new File(filepath);
		   FileInputStream  fis = new FileInputStream(file);

		      // Here BufferedInputStream is added for fast reading.
		   BufferedInputStream bis = new BufferedInputStream(fis);
		   DataInputStream dis = new DataInputStream(bis);

		      // dis.available() returns 0 if the file does not have more lines.
		      while (dis.available() != 0) {

		      // this statement reads the line from the file and print it to
		        // the console.
		    	  lastline = dis.readLine();
		        System.out.println(lastline);
		        
		      }

		      // dispose all the resources after using them.
		      fis.close();
		      bis.close();
		      dis.close();

		    } catch (FileNotFoundException e) {
		      e.printStackTrace();
		    } catch (IOException e) {
		      e.printStackTrace();
		    }
		    return lastline;
   }
   private void fileWrite(String filepath, String contents ) {
	   try{
		    // Create file 
		    FileWriter fstream = new FileWriter(filepath);
		        BufferedWriter out = new BufferedWriter(fstream);
		    out.write(contents);
		    //Close the output stream
		    out.close();
		    }catch (Exception e){//Catch exception if any
		      System.err.println("Error: " + e.getMessage());
		    }
   }
 
 
       
}
