package gov.nasa.gsfc.dqss;

import gov.nasa.gsfc.G4DataObject;
import gov.nasa.gsfc.Stopwatch;

import java.io.IOException;
import java.util.Iterator;
import java.util.List;

import ncsa.hdf.object.Dataset;
import ncsa.hdf.object.Group;

public class ModisMaskFactory extends MaskFactory {

	Stopwatch s;
	DataQuality dataQuality;
	
	public ModisMaskFactory(String dir, String configPath) {
		super();
		dqss = new DQSS(dir,configPath);
		/*
		 * This is where the DataQuality constructor should have been in the first
		 * place. Not in startMasking to reopen the ontology each time!
		 * We will always need a DataQuality instance but not necessarily one
		 * with it's own Ontology object.
		 */
		  dataQuality = new DataQuality();  
        /*
         * If we are not using the servlet for our ontology:
         */
        if (!DQSS.UseServlet) {
        	logger.warn("Calling dataQuality...Not expecting to use Servlet for Ontology");
        	dataQuality.setup(DQSS.ontology, DQSS.navigation);
        }
	}
	/*
	 * Queries for ontology stuff in various ways so this is not needed by
	 * Masker in MODAPS Arch. Just by the Web Services prep bit in MODAPS Arch
	 */ 
	 
	public void PrepMasking(DataFile file, String snvid, String dataVariable, 
			String criteria, DQSSNumber fillvalue){
		
		
		logger.warn("HDF File:" + file.getFileName());
        logger.warn("Dataset:" + snvid);
        logger.warn("DataVariable:" + dataVariable);
        logger.warn("Criteria:" + criteria);
        logger.warn("Staging Dir:" + DQSS.stagingDirectory);
        logger.warn("FillValue:" + fillvalue);
        
		  
	       //  Ontology Servlet Architecture:
	       
	    	boolean UseCache = DQSS.UseCache;
	    	 boolean cacheTheResults = false;
	        String OntologyURL = DQSS.OntologyCriteriaURL + "?";
	        OntologyURL += "dataset=" + snvid + "&" +
	                       "parameter=" + dataVariable + "&" +
	                       "criteria=" + criteria;
	        
	        logger.debug("OntologyURL:" + OntologyURL);
	        G4DataObject obj=null;
	        String address = OntologyURL;
	        if (UseCache) {
	        	 
	        	 if (DQSS.UseServlet) {
	        		 address = OntologyURL;
	        	 }
	        	 else {
	        		 address = DQSS.ontology + "?dataset=" + snvid + "&" +
	                 "parameter=" + dataVariable + "&" +
	                 "criteria=" + criteria;
	        	 }
	        	 logger.debug("USING CACHE");
	        	 
	        	  // Cache key is set here:
	        	  
	        	 obj = dataQuality.checkCache(address);
	        	 if (obj != null) {
	        		 logger.debug("We found cache key:" + obj.getKey());
	        	 }
	        }
	        else {
	        	logger.debug("NOT USING CACHE");
	        }
	       
	        
	        // This now encapsulates Servlet and local Jena algorithms.
	         
	    	try {
	    	 dataQuality.getOntologyResults(address,UseCache,obj,snvid,dataVariable,criteria);
				
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (Exception e) {
				e.printStackTrace();
			}
			
			
			s.stop();
		    logger.debug("Get Ontology Data : elapsed time: " + s.getElapsedTime());
		    s.reset();
		    s.start();
		   
		    /*
	        int status = dataQuality.XML2Castor();
	        if (status > 0) {
	           logger.error("XML Returned from Ontology does not match what castor was expecting for dataVariable:" + dataVariable);
	           return ;
	        }
	       */
	        
	        
	         //  Put ONTOLOGY DATA:
	         //  into file object
	         
	        //dataQuality.parseAirsOntologyResults((AIRSDataFile) file);
	        
		
	}

	
	
	public  Statistics StartMasking(DataFile file,   DQSSNumber fillvalue){
		
		s = new Stopwatch();
		s.start();
	
        /*
         * Not doing this for MODIS (two criteria for one reason)
       
		if (NoScreening(file.crit2num)) {
        	return null;
        }
        */
		
        Group newg   = file.getGroupOfDataset(dataVariable);
        /* 
         * Doing bits exercises
         */
        Dataset data = null;
       
	    data =   file.copyDatasetWithSuffix(dataVariable,newg, DQSS.MaskName );
	    
	    Dataset[] quality   = file.getQSVars();
	    /*
	     * Bit examination.
	     */
	    //examineQualityVariable(file,quality);
	    if (quality == null) {
	    	logger.fatal("Could not find any quality datasets for " + 
	    			dataVariable + " File " + file.getFileName() + " may not have been opened/exist");
	    	return null;
	    }
	    //file.NavigateSwathGroup() ;
	    FillValue = fillvalue;
	    DQSS.logger.debug("Using FillValue:" + fillvalue);
	    Mask mask = null;
	    if (quality != null) {
	    	logger.debug("I Found qs_var:" + quality[0].getName());
	    }
	    else {
	    	logger.error("I could not find a qs_var");
	    	return null;
	    }
	    if (data != null) {
	    	logger.debug("I Found dataVariable:" + data.getName());
	    }
	    else {
	    	logger.error("I could not find the dataVariable:" + dataVariable);
	    	return null;
	    }
	    DQSS.logger.debug(file.get_nConsRef() + ":" + file.get_nRef() + ":" + file.get_nCoordCons() + ":" + file.get_nQualLevel());
	    DQSS.logger.debug(file.nConsRef + ":" + file.nRef + ":" + file.nCoordCons + ":" + file.nQualLevel);
	    s.stop();
	    logger.debug("Get Ontology Data postprocessing: elapsed time: " + s.getElapsedTime());
	    s.reset();
	    s.start();
	    
	  
        if (file.nCoordCons == 0  && file.nQualLevel == 1) {
        	logger.warn("Doing " + dataVariable + "(rnk:" + data.getRank()  + ") in Modis1D1Q with " + file.qualityNames);
        	mask = new Modis1D1Q(data, quality, file, dir);
     
        }
        else if (file.nCoordCons == 0  && file.nQualLevel == 2) {
			
		     logger.warn("Doing " + dataVariable + "(rnk:" + data.getRank()  + ") in Modis1D2Q with " + file.qualityNames);
			 // Doing NIR       in Modis1D2Q 
    	     mask = new Modis1D2Q(data, quality, file, dir);

    	} 
		
		else  {
			logger.error("Could not find a MODIS case for masking for dataVariable:" + dataVariable);
			return null;
		}
		
		
		
        try {
				mask.doMask();
				logger.debug(data.getDimNames());
				 // doing in screener file.updateStructMetadata(dataVariable , DQSS.MaskName);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				logger.error("We don't have a MODIS masker for: " + dataVariable + " yet");
				return null;
			}
           	// Writing stats to the global metadata now. mask.writeStats();
           
        s.stop();
        logger.debug(mask.getClass().getName() + ", " + data.getName() 
        		+ " masking elapsed time: " + s.getElapsedTime());
        
       
		return mask.getStats();
         
	}
	

	public  String LoadOntology(String snvid, String dataVariable, 
			String criteria){

		
		logger.warn("Dataset:" + snvid);
        logger.warn("DataVariable:" + dataVariable);
        logger.warn("Criteria:" + criteria);
        logger.warn("Staging Dir:" + DQSS.stagingDirectory);
          /*
       *  Ontology Servlet Architecture:
       */
        String urlcriteria = handleMultipleCriteria(criteria);
        
    	boolean UseCache = DQSS.UseCache;
    	boolean cacheTheResults = false;
        String OntologyURL = DQSS.OntologyCriteriaURL + "?";
        OntologyURL += "dataset=" + snvid + "&" +
                       "parameter=" + dataVariable + "&" +
                        urlcriteria;
        
        logger.debug("OntologyURL:" + OntologyURL);
        G4DataObject obj=null;
        String address = OntologyURL;
        if (UseCache) {
        	 
        	 if (DQSS.UseServlet) {
        		 address = OntologyURL;
        	 }
        	 else {
        		 address = DQSS.ontology + "?dataset=" + snvid + "&" +
                 "parameter=" + dataVariable + "&" +
                  urlcriteria;
        	 }
        	 logger.debug("USING CACHE");
        	 /*
        	  * Cache key is set here:
        	  */
        	 obj = dataQuality.checkCache(address);
        	 if (obj != null) {
        		 logger.debug("We found cache key:" + obj.getKey());
        	 }
        }
        else {
        	logger.debug("NOT USING CACHE");
        }
       
        /*
         * This now encapsulates Servlet and local Jena algorithms.
         */
    	try {
    	 dataQuality.getOntologyResults(address,UseCache,obj,snvid,dataVariable,criteria);
			
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
		// don't need this with ts2/new version of servlet which includes datavar arrangexml(dataQuality.results);
		
		
	    return dataQuality.results;
	}
	
   
    
	boolean NoScreening(List<Long> criteria) {
		
		Iterator<Long> itr = criteria.iterator();
		Long skip = null;
		while (itr.hasNext()) {
			skip = itr.next();
			if (skip != DQSS.SkipBecauseNoScreeningSelected) {
				return false;
			}
		}
		if (skip == null) {
			logger.fatal("Criteria selected for this parameter was null");
			System.err.println("Criteria selected for this parameter was null" );
	        System.exit(1);
		}
		logger.warn("Skipping this parameter because no screening (" + skip + ") was selected");
		
		return true;
	}
	
	private String handleMultipleCriteria(String input) {
		String output = "";
		if (!input.contains(DQSS.delimiter)) {
			return "criteria=" + input;
		}
		String[] crits = input.split(DQSS.delimiter);
		for (int i = 0; i < crits.length; ++i) {
			output += "criteria=" + crits[i] + "&";
		}
		
		return output;
	}
	
}

