package gov.nasa.gsfc.dqss;

import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import ncsa.hdf.object.Dataset;

import org.apache.log4j.Logger;

/*
 * If the user increments 'kept' using inc() then he just needs 
 * to setTotal(i) at end .
 */
public class Statistics {

	int kept;
	int prefill;
	int total;
	int discarded;
	String directory;
	private String dataVariable;
	private String algorithm;
	private String filetype;
	//private String[] qualityLevel = new String[5];
	List<String> qualityLevel = new ArrayList<String>();
	//private String qualityVariables = "";
	List<String> qualityVariables = new ArrayList<String>();
	List<String> qualityVariableTypes = new ArrayList<String>();
	static Logger logger = Logger.getLogger(Statistics.class.getName());
	HashMap<Integer,Integer> cflags = new HashMap<Integer,Integer>(); 
	HashMap<Integer,Integer> qaflags = new HashMap<Integer,Integer>();  
	private long npixels;
	
	public Statistics(DataFile file) {
		filetype = file.FileType;
	}
	
	public Statistics(String dir, String param, DataFile file, Dataset[] QualityDatasets) {
		total = -1;
		kept = 0;
		discarded = -1;
		directory = dir;
		/* mask(name) is passed but we need the normal name in the
		 * global attributes.
		 */
		dataVariable = param.replace(DQSS.MaskName, "");
		
	    filetype = file.FileType;
	    logger.debug("Statistics constructor:" + dataVariable +
	    		":" + file.FileType); 
	    for (int i = 0; i < QualityDatasets.length; ++i) {
	    		qualityVariables.add( QualityDatasets[i].getName().trim());
	    		qualityVariableTypes.add(file.get_qs_types().get(i));
	    }
	}
	/*
	 * Don't expect this one to be used:
	 */
	public Statistics(String dir, String param) {
		total = -1;
		kept = 0;
		discarded = -1;
		directory = dir;
		dataVariable = param.trim();
	    
	}
	public String getAlgorithm() {
		return algorithm;
	}
	public int getTotal() {
		return total;
	}
	
	public int getKept() {
		return kept;
	}
	
	public void setQualityLevel(long level, int ith, String name) {
		logger.debug(name + ":" + level);
		if (filetype.contains("AIRS")) {
			switch((int) level) {
			case 0: qualityLevel.add("Best");
			break;
			case 1: qualityLevel.add("Good");
			break;
			case 2: qualityLevel.add("DoNotUse");
			break;
			default: ;
			}
				
		}
		else if (filetype.contains("MOD")) {
			if (name.contains("Cloud")) {
				switch((int) level) {
				case 0: qualityLevel.add(DQSS.Q2String0);
				break;
				case 1: qualityLevel.add(DQSS.Q2String1);
				break;
				case 2: qualityLevel.add(DQSS.Q2String2);
				break;
				case 3: qualityLevel.add(DQSS.Q2String3);
				default: ;
				}
			}
			else {
				switch((int) level) {
				case 0: qualityLevel.add(DQSS.Q1String0);
				break;
				case 1: qualityLevel.add(DQSS.Q1String1);
				break;
				case 2: qualityLevel.add(DQSS.Q1String2);
				break;
				case 3: qualityLevel.add(DQSS.Q1String3);
				default: ;
				}
			}
		}
	}

	public String getQualityLevel(int ith) {
		if (this.qualityLevel.size() >= ith + 1) {
			return this.qualityLevel.get(ith);
		}
		return null;
	}
	public int getDiscarded() {
		return discarded;
	}
	public void setTotal(int total) {
		 this.total = total;
		 /*
		 if (kept > 0) {
			 setDiscarded(total - kept);
		 }
		 */
	}
	public void setKept(int kept) {
		 this.kept = kept;
		 /*
		 if (total > 0) {
			 setDiscarded(total - kept);
		 }
		 */
	}
	
	public void setDiscarded(int discarded) {
		 this.discarded = discarded;
	}
	
	/*
	 * Number of pixels set to 1 in mask
	 */
	public void inc() {
		++kept;
	}
	
	/* 
	 * Number of pixels set to fill value before masking
	 */
	public void countfill() {
		++prefill;
	}
	
	public String getVariable() {
		return dataVariable;
	}
	public String getQualityVariable(int ith) {
		return this.qualityVariables.get(ith);
	}
	public String getQualityVariableType(int ith) {
		return this.qualityVariableTypes.get(ith);
	}
	/*
	 * Percentage of pixels in mask set to 1
	 */
	public String getPercentage() {
		 float n = this.kept;
		 float d = this.total;
		 float percentage = (float) (100.0 * (n/d));
		 logger.debug("kept:" + n + " total:" + d + " prefill:" + prefill);
		 return new DecimalFormat("##.##").format(percentage);
	}
	
	/*
	 * percentage of non-fill values before mask.
	 */
	public String getPrePercentage() {
		 float n = this.prefill;
		 float d = this.total;
		 float percentage = (float)( 100 - (100.0 * (n/d)));
		
		 return new DecimalFormat("##.##").format(percentage);
		 
	}
	
	
	 public boolean writeStats() {
		 String percentage = getPercentage();	 
		 String output = dataVariable  + " mask retained " +
		 kept + " of " + total + " pixels. " + percentage + "%";
		 //String currentcontents = fileRead();
		 return DQSS.fileAppend(output);
	 }
	 /*
	  * just integers for flags
	  */
	 public void countCldflags(int value) 
	 {
		 Integer currentTotalForValue = this.cflags.get(value);
		 if (currentTotalForValue == null) {
			 currentTotalForValue = 0;
		 }
		 currentTotalForValue += 1;
		 this.cflags.put(value,currentTotalForValue);
	 }
	 public void countQAflags(int value) 
	 {
		 Integer currentTotalForValue = this.qaflags.get(value);
		 if (currentTotalForValue == null) {
			 currentTotalForValue = 0;
		 }
		 currentTotalForValue += 1;
		 this.qaflags.put(value,currentTotalForValue);
	 }
	 public void setpixels(long rows, long cols) {
		 npixels = rows * cols;
	 }
	 public void publishCldFlags() {
		
		 logger.debug("Publish Cld Flags");
		 Iterator it = this.cflags.entrySet().iterator();
		 long total = 0;
		 while (it.hasNext()) {
              Map.Entry pairs = (Map.Entry)it.next();
              Integer key = (Integer)pairs.getKey();
              Integer value = (Integer)pairs.getValue();
              logger.warn("Flags:" + key + " Value:" + value);
              total += value;
        }
		 logger.warn("Total C Flags Pixels:" + total + "(" + npixels + ")");
	 }
	 public void publishQAFlags() {
			
		 logger.debug("Publish QA Flags");
		 Iterator it = this.qaflags.entrySet().iterator();
		 long total = 0;
		 while (it.hasNext()) {
              Map.Entry pairs = (Map.Entry)it.next();
              Integer key = (Integer)pairs.getKey();
              Integer value = (Integer)pairs.getValue();
              logger.warn("Flags:" + key + " Value:" + value);
              total += value;
        }
		 logger.warn("Total QA Flags Pixels:" + total + "(" + npixels + ")");
	 }
		
}
